﻿using System;
using System.Globalization;

namespace FormulaEnginePoC.DynamicExpressionParse
{
    internal class ExpressionParserBase
    {
        public struct Token
        {
            public TokenId Id;
            public string Text;
            public int Pos;
        }

        public enum TokenId
        {
            // ReSharper disable once UnusedMember.Global
            Unknown,

            End,
            Identifier,
            StringLiteral,
            IntegerLiteral,
            RealLiteral,
            Exclamation,
            Percent,
            Amphersand,
            OpenParen,
            CloseParen,
            Asterisk,
            Plus,
            Comma,
            Minus,
            Dot,
            Slash,
            Colon,
            LessThan,
            Equal,
            GreaterThan,
            Question,
            OpenBracket,
            CloseBracket,
            Bar,
            ExclamationEqual,
            DoubleAmphersand,
            LessThanEqual,
            LessGreater,
            DoubleEqual,
            GreaterThanEqual,
            DoubleBar
        }

        private readonly string _text;
        private char _ch;

        protected int TextPos;
        protected readonly int TextLen;
        protected Token CurrentToken;

        protected ExpressionParserBase(string expression)
        {
            _text = expression ?? throw new ArgumentNullException(nameof(expression));
            TextLen = _text.Length;

            SetTextPos(0);
        }

        private void NextChar()
        {
            if (TextPos < TextLen) TextPos++;
            _ch = TextPos < TextLen ? _text[TextPos] : '\0';
        }

        protected void SetTextPos(int pos)
        {
            TextPos = pos;
            _ch = TextPos < TextLen ? _text[TextPos] : '\0';
        }

        private void ValidateDigit()
        {
            if (!char.IsDigit(_ch)) throw ParseError(TextPos, Res.DigitExpected);
        }

        protected Exception ParseError(string format, params object[] args)
        {
            return ParseError(CurrentToken.Pos, format, args);
        }

        protected Exception ParseError(int pos, string format, params object[] args)
        {
            return new ParseException(string.Format(CultureInfo.CurrentCulture, format, args), pos);
        }

        protected void NextToken()
        {
            while (char.IsWhiteSpace(_ch)) NextChar();
            var tokenPos = TextPos;

            if (false == ParseToken(out var t))
            {
                throw ParseError(TextPos, Res.InvalidCharacter, _ch);
            }

            CurrentToken.Id = t;
            CurrentToken.Text = _text.Substring(tokenPos, TextPos - tokenPos);
            CurrentToken.Pos = tokenPos;
        }

        private bool ParseToken(out TokenId t)
        {
            switch (_ch)
            {
                case '!':
                    NextChar();
                    if (_ch == '=')
                    {
                        NextChar();
                        t = TokenId.ExclamationEqual;
                    }
                    else
                    {
                        t = TokenId.Exclamation;
                    }

                    break;

                case '%':
                    NextChar();
                    t = TokenId.Percent;
                    break;

                case '&':
                    NextChar();
                    if (_ch == '&')
                    {
                        NextChar();
                        t = TokenId.DoubleAmphersand;
                    }
                    else
                    {
                        t = TokenId.Amphersand;
                    }

                    break;

                case '(':
                    NextChar();
                    t = TokenId.OpenParen;
                    break;

                case ')':
                    NextChar();
                    t = TokenId.CloseParen;
                    break;

                case '*':
                    NextChar();
                    t = TokenId.Asterisk;
                    break;

                case '+':
                    NextChar();
                    t = TokenId.Plus;
                    break;

                case ',':
                    NextChar();
                    t = TokenId.Comma;
                    break;

                case '-':
                    NextChar();
                    t = TokenId.Minus;
                    break;

                case '.':
                    NextChar();
                    t = TokenId.Dot;
                    break;

                case '/':
                    NextChar();
                    t = TokenId.Slash;
                    break;

                case ':':
                    NextChar();
                    t = TokenId.Colon;
                    break;

                case '<':
                    NextChar();
                    if (_ch == '=')
                    {
                        NextChar();
                        t = TokenId.LessThanEqual;
                    }
                    else if (_ch == '>')
                    {
                        NextChar();
                        t = TokenId.LessGreater;
                    }
                    else
                    {
                        t = TokenId.LessThan;
                    }

                    break;

                case '=':
                    NextChar();
                    if (_ch == '=')
                    {
                        NextChar();
                        t = TokenId.DoubleEqual;
                    }
                    else
                    {
                        t = TokenId.Equal;
                    }

                    break;

                case '>':
                    NextChar();
                    if (_ch == '=')
                    {
                        NextChar();
                        t = TokenId.GreaterThanEqual;
                    }
                    else
                    {
                        t = TokenId.GreaterThan;
                    }

                    break;

                case '?':
                    NextChar();
                    t = TokenId.Question;
                    break;

                case '[':
                    NextChar();
                    t = TokenId.OpenBracket;
                    break;

                case ']':
                    NextChar();
                    t = TokenId.CloseBracket;
                    break;

                case '|':
                    NextChar();
                    if (_ch == '|')
                    {
                        NextChar();
                        t = TokenId.DoubleBar;
                    }
                    else
                    {
                        t = TokenId.Bar;
                    }

                    break;

                case '"':
                case '\'':
                    var quote = _ch;
                    do
                    {
                        NextChar();
                        while (TextPos < TextLen && _ch != quote) NextChar();
                        if (TextPos == TextLen)
                            throw ParseError(TextPos, Res.UnterminatedStringLiteral);
                        NextChar();
                    } while (_ch == quote);

                    t = TokenId.StringLiteral;
                    break;

                default:
                    if (char.IsLetter(_ch) || _ch == '@' || _ch == '_')
                    {
                        do
                        {
                            NextChar();
                        } while (char.IsLetterOrDigit(_ch) || _ch == '_');

                        t = TokenId.Identifier;
                        break;
                    }

                    if (char.IsDigit(_ch))
                    {
                        t = TokenId.IntegerLiteral;
                        do
                        {
                            NextChar();
                        } while (char.IsDigit(_ch));

                        if (_ch == '.')
                        {
                            t = TokenId.RealLiteral;
                            NextChar();
                            ValidateDigit();
                            do
                            {
                                NextChar();
                            } while (char.IsDigit(_ch));
                        }

                        if (_ch == 'E' || _ch == 'e')
                        {
                            t = TokenId.RealLiteral;
                            NextChar();
                            if (_ch == '+' || _ch == '-') NextChar();
                            ValidateDigit();
                            do
                            {
                                NextChar();
                            } while (char.IsDigit(_ch));
                        }

                        if (_ch == 'F' || _ch == 'f') NextChar();
                        break;
                    }

                    if (TextPos == TextLen)
                    {
                        t = TokenId.End;
                        break;
                    }

                    t = TokenId.Unknown;
                    return false;
            }

            return true;
        }
    }
}